use crate::config::Config;
use crate::sites::SitesWorker;
use camino::Utf8Path;
use color_eyre::eyre::{WrapErr, eyre};
use serde::Deserialize;
use std::sync::Arc;
use async_std::{fs, task};
use async_std::task::JoinHandle;
use tide::{Request, StatusCode};

#[derive(Clone)]
struct State {
    sites_worker: Arc<SitesWorker>,
    config: &'static Config,
}

#[derive(Deserialize)]
struct NotifyBody {
    pub id: String,
    pub download_url: String,
}

pub async fn start_api_server(
    api_socket_path: &Utf8Path,
    sites_worker: Arc<SitesWorker>,
    config: &'static Config,
) -> color_eyre::Result<JoinHandle<color_eyre::Result<!>>> {
    let mut server = tide::with_state(State { sites_worker, config });

    server.at("/notify").post(handle_notify);

    server.at("/version").get(|_| async { Ok(env!("CARGO_PKG_VERSION")) });

    let path = api_socket_path.to_path_buf().into_std_path_buf();
    let _ = fs::remove_file(&path).await;

    Ok(task::spawn(async {
        server.listen(path).await.wrap_err("IO error in API server")?;
        Err(eyre!("The API unexpectedly stopped without error."))
    }))
}

async fn handle_notify(mut request: Request<State>) -> tide::Result {
    let provided_secret = request
        .header("Authorization")
        .and_then(|h| h.get(0))
        .and_then(|h| h.to_string().strip_prefix(&"Bearer ").map(|s| s.to_owned()))
        .ok_or(tide::Error::from_str(StatusCode::Unauthorized, "No secret was provided."))?;

    let domain = request.host().unwrap().to_string();

    let (_, scope) = request
        .state()
        .config
        .scopes
        .iter()
        .find(|(_, s)| s.secret == provided_secret)
        .ok_or(tide::Error::from_str(StatusCode::Unauthorized, "The provided secret is invalid."))?;

    if scope.domain_pattern.is_match(&domain) {
        let body = request.body_json::<NotifyBody>().await?;
        let sites_worker = Arc::clone(&request.state().sites_worker);

        sites_worker.set_current_version(domain, body.id, body.download_url).await.unwrap();

        Ok(StatusCode::NoContent.into())
    } else {
        Err(tide::Error::from_str(
            StatusCode::Unauthorized,
            "The provided secret is not valid for this domain.",
        ))
    }
}
