#![feature(never_type)]
#![feature(duration_constructors)]
#![feature(pattern)]

use crate::api::start_api_server;
use crate::caddy::start_caddy;
use crate::config::load_config;
use crate::sites::start_sites_worker;
use camino::Utf8Path;
use color_eyre::Result;
use color_eyre::eyre::WrapErr;
use log::LevelFilter;
use std::sync::Arc;
use async_std::fs;

mod api;
mod caddy;
mod config;
mod sites;

#[async_std::main]
async fn main() -> Result<()> {
    color_eyre::install()?;
    env_logger::Builder::new().filter_module("sscdc", LevelFilter::Info).parse_default_env().init();

    log::info!("Loading configuration…");
    let config = load_config().await?;

    let sockets_directory_path = Utf8Path::new(&config.sockets_directory);
    fs::create_dir_all(sockets_directory_path.as_std_path()).await.wrap_err("Failed to access or create the sockets directory.")?;
    let sockets_directory_path = sockets_directory_path.canonicalize_utf8().unwrap();
    let api_socket_path = sockets_directory_path.join("api.sock");

    log::info!("Starting the reverse proxy…");
    let caddy_controller = start_caddy(&api_socket_path, &sockets_directory_path, config.http_port).await?;

    let (sites_worker, sites_worker_join_handle) = start_sites_worker(config.sites_directory.to_string().into(), caddy_controller).await?;

    log::info!("Starting internal API server…");
    start_api_server(&api_socket_path, Arc::new(sites_worker), Box::leak(Box::new(config))).await?;
    log::debug!("The internal API server is listening at {api_socket_path}");

    log::info!("Startup complete.");
    sites_worker_join_handle.await?;
}
